\ 4tH deskcalculator - (c) 1997,9 HanSoft & Partners

false constant ANS                     \ if true, we compile for ANS-Forth

ANS [IF]
1024 constant /STRING-SPACE            \ size of String Segment
s" easy4th.f" included                 \ load interface
: length$ dup c@ ;                     \ Forth compatible count
[THEN]

ANS [NOT] [IF]
: length$ count ;                      \ 4tH compatible count
[THEN]

\ In this section the exceptions are defined
\ and several routines that handle them

1 constant #UndefName
2 constant #BadVariable

8 constant #variables

: variable?                            \ is it a valid variable?
  dup                                  ( n n)
  #variables 1- invert and             ( n f)
  if
    #BadVariable throw
  then
;

\ Here all internal commands are listed
\ and placed inside the dictionary.
\ Checking is done by the routines above.
\ Note: there are aliases!

#variables array variables

: +_      + ;                          \ + command
: -_      - ;                          \ - command
: *_      * ;                          \ * command
: /_      / ;                          \ / command
: ._      . ;                          \ . command
: .r_     .r ;                         \ .r command
: drop_   drop ;                       \ drop command
: dup_    dup ;                        \ dup command
: rot_    rot ;                        \ rot command
: /mod_   /mod ;                       \ /mod command
: */_     */ ;                         \ */ command
: */mod_  */mod ;                      \ */mod command
: swap_   swap ;                       \ swap command
: over_   over ;                       \ over command
: !_      variable? variables + ! ;    \ ! command
: +!_     variable? variables + +! ;   \ +! command
: @_      variable? variables + @ ;    \ @ command
: ?_      variable? variables + ? ;    \ ? command
: mod_    mod ;                        \ mod command
: abs_    abs ;                        \ abs command
: negate_ negate ;                     \ negate command
: invert_ invert ;                     \ invert command
: min_    min ;                        \ min command
: max_    max ;                        \ max command
: or_     or ;                         \ or command
: xor_    xor ;                        \ xor command
: and_    and ;                        \ and command
: lshift_ lshift ;                     \ lshift command
: rshift_ rshift ;                     \ rshift command
: inc     1+ ;                         \ inc command
: dec     1- ;                         \ dec command
: 2*_     2* ;                         \ 2* command
: 2/_     2/ ;                         \ 2/ command
: spaces_ spaces ;                     \ spaces command
: wait_   wait ;                       \ wait command
: emit_   emit ;                       \ emit command
: base!   36 min 2 max base ! ;        \ base command
: decimal_ decimal ;                   \ decimal command
: hex_  hex ;                          \ hex command
: octal_ octal ;                       \ octal command
: binary 2 base! ;                     \ binary command
: depth_ depth ;                       \ depth command
: bye quit ;                           \ quit command
: dummy ;                              \ dummy command
: space_ space ;                       \ space command
: random_ random ;                     \ random command
: cr_ cr ;                             \ cr command
: A. 0 ;                               \ variable A
: B. 1 ;                               \ variable B
: C. 2 ;                               \ variable C
: D. 3 ;                               \ variable D
: E. 4 ;                               \ variable E
: F. 5 ;                               \ variable F
: G. 6 ;                               \ variable G
: H. 7 ;                               \ variable H
: .(_ 41 word count type ;             \ .( command
: (_ 41 word drop ;                    \ ( command
: char_ bl word c@ ;                   \ char command

ANS [IF]                               \ time not available in ANS-Forth
: time_ ." Unavailable" cr ;
[THEN]

ANS [NOT] [IF]
: time_ time ;                         \ time command
[THEN]

-1 constant NULL                       \ NULL pointer

create dictionary                      \ dictionary
  " +"       , ' +_       ,
  " th"      , ' +_       ,
  " -"       , ' -_       ,
  " *"       , ' *_       ,
  " /"       , ' /_       ,
  " quit"    , ' bye      ,
  " bye"     , ' bye      ,
  " q"       , ' bye      ,
  " ."       , ' ._       ,
  " .r"      , ' .r_      ,
  " drop"    , ' drop_    ,
  " dup"     , ' dup_     ,
  " rot"     , ' rot_     ,
  " swap"    , ' swap_    ,
  " over"    , ' over_    ,
  " A."      , ' A.       ,
  " B."      , ' B.       ,
  " C."      , ' C.       ,
  " D."      , ' D.       ,
  " E."      , ' E.       ,
  " F."      , ' F.       ,
  " G."      , ' G.       ,
  " H."      , ' H.       ,
  " !"       , ' !_       ,
  " +!"      , ' +!_      ,
  " @"       , ' @_       ,
  " ?"       , ' ?_       ,
  " base!"   , ' base!    ,
  " decimal" , ' decimal_ ,
  " octal"   , ' octal_   ,
  " hex"     , ' hex_     ,
  " binary"  , ' binary   ,
  " .("      , ' .(_      ,
  " mod"     , ' mod_     ,
  " abs"     , ' abs_     ,
  " negate"  , ' negate_  ,
  " invert"  , ' invert_  ,
  " min"     , ' min_     ,
  " max"     , ' max_     ,
  " or"      , ' or_      ,
  " and"     , ' and_     ,
  " xor"     , ' xor_     ,
  " lshift"  , ' lshift_  ,
  " rshift"  , ' rshift_  ,
  " depth"   , ' depth_   ,
  " cells"   , ' dummy    ,
  " 1+"      , ' inc      ,
  " cell+"   , ' inc      ,
  " 1-"      , ' dec      ,
  " cell-"   , ' dec      ,
  " space"   , ' space_   ,
  " spaces"  , ' spaces_  ,
  " random"  , ' random_  ,
  " wait"    , ' wait_    ,
  " 2*"      , ' 2*_      ,
  " 2/"      , ' 2/_      ,
  " emit"    , ' emit_    ,
  " char"    , ' char_    ,
  " [char]"  , ' char_    ,
  " time"    , ' time_    ,
  " ("       , ' (_       ,
  " /mod"    , ' /mod_    ,
  " */"      , ' */_      ,
  " */mod"   , ' */mod_   ,
  " cr"      , ' cr_      ,
  NULL ,

\ This is the standard interpreter as presented in
\ the library routines. It is unaltered except that
\ exceptions are used and some sections have been
\ moved.

/pad string buffer                     \ temporary space for string
true constant ?ignore                  \ ignore case
                                       \ this routine compares two strings
: match                                ( a1 a2 -- f )
  length$ 1+                           ( a1 a2 n )
  dup 0                                ( a1 a2 n n 0 )
  do                                   ( a1 a2 f )
    drop                               ( a1 a2 )
    over i + c@                        ( a1 a2 c1)
    over i + c@                        ( a1 a2 c1 c2)
    xor ?ignore                        ( a1 a2 n f)
    if                                 ( a1 a2 n)
      32 invert and                    ( a1 a2 f)
    then                               ( a1 a2 f)
    dup                                ( a1 a2 f f)
    if                                 ( a1 a2 f)
      leave
    then                               ( a1 a2 f)
  loop
  >r drop drop r>                      ( f )
;
                                       \ this routine finds the matching token
: find                                 ( a -- xt f | a -f )
  dictionary                           ( a1 tk )

  begin
    over over                          ( a1 tk a1 tk)
    @' dup NULL <>                     ( a1 tk a1 s f )
    if                                 ( a1 tk a1 s )
      buffer copy                      ( a1 tk a1 a2 )
      match dup                        ( a1 tk f2 f1 )
    else
      drop drop true false             ( a1 tk f2 f1 )
    then
  while                                ( a1 tk f2 )
    drop                               ( a1 tk )
    cell+ cell+                        ( a1 tk )
  repeat                               ( a1 tk f )

  if                                   ( a1 tk )
    drop false                         ( a1 f )
  else
    swap drop                          ( tk )
    cell+ @' true                      ( xt f )
  then
;
                                       \ this routine interprets a string
: interpret                            ( -- )
  begin
    bl word length$ 0<>                ( a f )
  while                                ( a )
    find                               ( xt f | a -f )
    if                                 ( xt )
      execute                          ( -- )
    else                               ( a )
      number                           ( n )
      dup (error) =                    ( n f )
      if                               ( n )
        drop                           ( -- )
        #UndefName throw
      then
    then
  repeat                               ( a )
  drop                                 ( -- )
;

\ This is the actual DC program. It features only a
\ message handler and a interpretation-loop.
\ The rest is handled by library routines.

create Message
  " Exception ignored" ,
  " Undefined name" ,
  " Bad variable" ,

: ShowMessage                          \ show error message
  0 max Message swap th @' buffer copy ( n --)
  count type space                     \ print message
;

: dc
  begin
    ." OK" cr
    refill drop
    ['] interpret
    catch dup
    if
      ShowMessage
    else
      drop
    then
  again
;

ANS [NOT] [IF] dc [THEN]
