/*
+-------+-----------------+------+-----------------+-------+-----------------+
|Program| EXEC_4TH.C      |Header| 4TH.H           |Library| CEXT            |
+-------+-----------------+------+-----------------+-------+----------+------+
|HISTORY                                           |    Programmer    | Date |
+--------------------------------------------------+------------------+------+
| Turbo C V2.0 version                             | HanSoft          |181094|
| DPX/2 340 version                                | HanSoft          |181094|
| Coherent V4.2 version                            | HanSoft          |201094|
+--------------------------------------------------+------------------+------+
|MODIFICATION LOG                                  |    Programmer    | Date |
+--------------------------------------------------+------------------+------+
| Added /MOD, * /, * /MOD, I', J, FAST, SLOW       | HanSoft          |191094|
| Removed Flags, changed NOP to NOOP               | HanSoft          |191094|
| Added M4UNKKEY message, .R, VARIABLE, V0@, USER@ | HanSoft          |201094|
| Added STACK@, LIMIT@, BASE@, SP@, RP@            | HanSoft          |201094|
| Added ABORT, EXIT, EXECUTE, TICK, C@, WAIT       | HanSoft          |241094|
| Changed allocation of RStack, Stack, Vars        | HanSoft          |241094|
| Added check on I/O, /. and /.R                   | HanSoft          |271094|
| Added randomizer, RANDOM and TIME                | HanSoft          |161194|
| Optimized +LOOP a little; fixed LEAVE bug        | HanSoft          |140395|
| Optimized +LOOP a little more                    | HanSoft          |150395|
| Fixed bug: ErrNo4th returns wrong errorcode      | HanSoft          |160395|
| Changed , into C, and removed V0@                | HanSoft          |180395|
| Switched LIMIT@ and USER@ commands               | HanSoft          |180395|
| Merged Stack and RStack                          | HanSoft          |180395|
| Fixed bug in ABORT                               | HanSoft          |210395|
| Rearranged commands to optimize performance      | HanSoft          |210395|
| Changed LIMIT@ to VARS, USER@ to USER, C, to ,   | HanSoft          |310395|
| Changed STACK@ to STACK; added C,                | HanSoft          |310395|
| Produced first interpreter of 4tH V2.0           | HanSoft          |200495|
| Redesigned allocation of variables               | HanSoft          |240495|
| Removed STACK and VALUE                          | HanSoft          |240495|
| Reduced field in .R and /.R to max. 31 positions | HanSoft          |250495|
| Removed obsolete check reading in values         | HanSoft          |250495|
| Made +LOOP Forth-compatible again                | HanSoft          |280495|
| Removed INPUT, added string words                | HanSoft          |280495|
| Added INRSET, changed PAD to FORMAT on some words| HanSoft          |290495|
| Increased field in .R and /.R to max. 63 pos.    | HanSoft          |300495|
| Changed CONSTANT to LITERAL                      | HanSoft          |300495|
| Added #, #S, <#, #>, HOLD & SIGN; removed /., /.R| HanSoft          |010595|
| Changed ,@; added COPY                           | HanSoft          |020595|
| Added sharpS(), sign(), lsharp(), type(), dotR() | HanSoft          |050995|
| Removed dot(), changed emit(); removed BASEFCH   | HanSoft          |050995|
| All I/O passes through emit(); added *Base       | HanSoft          |050995|
| Made code more solid; added M4BADRDX messages    | HanSoft          |060995|
| Fixed bug in dotR() and RSHARP code              | HanSoft          |060995|
| Changed ,@ to '@; set radix limits correctly     | HanSoft          |060995|
| Removed LEFT, RIGHT, !IN; added SHIFT and *In    | HanSoft          |110995|
| Rewrote TEXT completely                          | HanSoft          |120995|
| Fixed tiny bugs: init >IN, loading extern vars   | HanSoft          |130995|
| Added EXPECT command                             | HanSoft          |140995|
| Added expect(), FILE>, TTY>, DevIn               | HanSoft          |150995|
| Added FILE* to expect(); QUERY reads from TTY    | HanSoft          |160995|
| Added INFILE and OUTFILE                         | HanSoft          |180995|
| Fixed small bug in TEXT (incrementing >IN)       | HanSoft          |200995|
| Updated errors                                   | HanSoft          |220995|
| Changed EXPECT to ACCEPT, TEXT to WORD           | HanSoft          |080296|
| Added LOOP; removed conditionals                 | HanSoft          |080296|
| Changed WORD, NUMBER and PRINT                   | HanSoft          |080296|
| Removed RETURN                                   | HanSoft          |090296|
| Added cell support; changed COPY; added NE0      | HanSoft          |100296|
| Added Hcode support                              | HanSoft          |120296|
| Added HI and support for allocation              | HanSoft          |140296|
| Added extra check to NUMBER                      | HanSoft          |160296|
| Added TO and VALUE                               | HanSoft          |240296|
| Changed NOT4 to INVERT                           | HanSoft          |040396|
| Removed LINE; changed VARS to LAST, USER to FIRST| HanSoft          |050396|
| Added return-value to pop() and removed argument | HanSoft          |210496|
| Cleaned up code and added comment                | HanSoft          |220496|
| Changed store(), fetch(), sharp()                | HanSoft          |220496|
| Added additional optimizations in pop() calls    | HanSoft          |220496|
| Added some additional casts                      | HanSoft          |230496|
| Removed some casts                               | HanSoft          |060596|
| Rewrote to accomodate new Hcode format           | HanSoft          |021096|
| Added RSP and SP                                 | HanSoft          |041096|
| Renamed TICKFCH to FETCHPRM ('@ changed to @')   | HanSoft          |101096|
| Moved RSP and SP to headerfile                   | HanSoft          |201196|
| Removed ABORT, added SP! and RP!                 | HanSoft          |201196|
| Replaced M4UNKTOK by M4BADOBJ; added F_F and F_T | HanSoft          |031296|
| Added CLOSEIO                                    | HanSoft          |061296|
| Removed TTY>, >TTY, FILE>, >FILE, INFILE, OUTFILE| HanSoft          |280297|
| Added TTY, FILE, OPEN, CLOSE; removed CLOSEIO    | HanSoft          |280297|
| Optimized TRAILING a little                      | HanSoft          |150397|
| Made In local, added RESULT variable             | HanSoft          |100497|
| Added )                                          | HanSoft          |110497|
| Added throw(), THROW and HANDLER variable        | HanSoft          |170497|
| Added PassThru to throw()                        | HanSoft          |180497|
| Added M4THROW                                    | HanSoft          |190497|
| Optimized throw() a little                       | HanSoft          |200497|
| Changed return of rpop() on error                | HanSoft          |300497|
| Changed return of pop() on error                 | HanSoft          |040597|
| Added LOCAL_H support; fixed GCC warning (Now)   | HanSoft          |220597|
| Simplified PRINT and COPY; added toCharSeg()     | HanSoft          |240797|
| Added ARGN and ARGS                              | HanSoft          |240797|
| Fixed small bug in ARGS; cleaned up COPY         | HanSoft          |250797|
| Removed <limits.h>; fixed serious bug in throw() | HanSoft          |250797|
| Removed some redundant Applet->ErrNo             | HanSoft          |260797|
| Fixed GCC warning (setjmp); fixed small bug      | HanSoft          |270797|
| Changed throw(), pop() and rpop()                | HanSoft          |300797|
| Added BRANCHXT; changed header fieldname         | HanSoft          |120997|
| Removed BRANCHXT; replaced some variables        | HanSoft          |170997|
| Removed RCOPY2 and SHOW                          | HanSoft          |260997|
| Added unit to make Character Segment unsigned    | HanSoft          |270997|
| Removed INC2 and DEC2; added QDO                 | HanSoft          |151097|
| Removed QUERY, added REFILL                      | HanSoft          |311097|
| Fixed DJGPP I/O error in accept() when EOT       | HanSoft          |061197|
| Added CATCH and CAUGHT, changed Throw()          | HanSoft          |191197|
| Changed RANDOM to use RNDMASK                    | HanSoft          |221197|
| Removed EasyC syntax                             | HanSoft          |150198|
| Changed fileIO() for new file attributes         | HanSoft          |160298|
| Updated version info                             | HanSoft          |180298|
| Restored fileIO() and redesigned FOPEN code      | HanSoft          |200298|
| Added SQUOTE                                     | HanSoft          |200298|
| Integrated Web4tH variant                        | HanSoft          |290999|
| Removed unit datatype                            | HanSoft          |301099|
| Fixed error in accept() for GCC (flushing)       | HanSoft          |200600|
| Restored accept(); REFILL works again            | HanSoft          |200600|
| Removed WORD; added TOKEN and SKIP               | HanSoft          |160201|
| Fixed small bug in FCLOSE; redirect to tty       | HanSoft          |240401|
| Removed fileIO(); added OpenDevice, CloseDevice()| HanSoft          |230202|
| Fixed small bug in CloseDevice(), forced fclose()| HanSoft          |230202|
| Cleaned up pipe support for non-Unix platforms   | HanSoft          |240202|
| Pipes on non-Unix platforms always fail          | HanSoft          |240202|
| Added USE, removed TTY and FILE                  | HanSoft          |280202|
| Removed OpenDevice(), CloseDevice()              | HanSoft          |280202|
| Added OpenStream(), CloseStream()                | HanSoft          |280202|
| Changed emit(), accept()                         | HanSoft          |280202|
| Added NOPIPES directive; pipes are buggy         | HanSoft          |070302|
| Added fix for pipes; Popen() and Pclose()        | HanSoft          |080302|
| Fixed EOF bug in accept()                        | HanSoft          |090302|
| Restricted <dos.h> to TurboC to accomodate LCCW32| HanSoft          |060502|
| Added static modifier to Popen() and Pclose()    | HanSoft          |170902|
+--------------------------------------------------+------------------+------+
|REMARKS                                           |    Programmer    | Date |
+--------------------------------------------------+------------------+------+
| None                                             |                  |      |
+--------------------------------------------------+------------------+------+

  Copyright (C) 1994,2002 J.L. Bezemer

  This file is part of 4tH

  4tH is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


This program executes a 4tH program. The user variables can be preloaded and
used inside a 4th program. VarN is assigned to the number of variables
that are preloaded. ArgN is equivalent to argc; ArgS is equivalent to argv.
This interface allows you to pass arguments and strings to a 4tH program.
It returns the final value on the stack.

Prototype: cell exec_4th (Hcode *Object, unsigned ArgN, char **ArgS,
           unsigned VarN, ...)
Returns  : cell, final value on the stack on succes or CELL_MIN on error
Memory   : the size of the stack and variables and the user area
Example  :


#include "4th.h"
#include <stdlib.h>
#include <ctype.h>

#ifndef ARCHAIC
  int main (int argc, char **argv)
#else
  int main (argc, argv) int argc; char **argv;
#endif

{
  Hcode *object;
  char  *sources;
  cell   result;

  if (argc >= 2)
     {
       if ((sources = open_4th (argv [1])) == NULL)
          printf ("Loading; \'%s\' does not exist or too large\n", argv [1]);

       else
         {
           object = comp_4th (sources);

           if (object)
             {
               result = exec_4th (object, argc - 1, argv + 1, 2, 125L, 6L);
               printf ("Execute; word %u: %s\n", object->ErrLine,
                      errs_4th [object->ErrNo]);
               printf ("\nResult of script: %ld\n", result);
               return (EXIT_SUCCESS);
             }
         }
     }
  return (EXIT_FAILURE);
}


Input    : None
Output   : depends on program
Related  : dump_4th(), comp_4th()
*/


#ifdef USRLIB4TH
#include <sys/cmds_4th.h>
#include <4th.h>
#else
#include "cmds_4th.h"
#include "4th.h"
#endif

#include <ctype.h>
#include <stdlib.h>
#include <setjmp.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>

#ifdef UNIX
#include <sys/wait.h>
#endif

#ifdef __TURBOC__
#include <dos.h>
#else
#include <unistd.h>
#include <errno.h>
#endif

#define EOTu   4                       /* Unix end of text character */
#define EOTm  26                       /* Microsoft end of text character */

#define EXCEPTION 1                    /* label for longjmp() */
#define SEED ((unsigned)(time(&Now)%37L))
                                       /* randomizer seed */
#define CIN  ((int)F4_READ)            /* input device number */
#define COUT ((int)F4_WRITE)           /* output device number */
                                       /* Unix pipe support */
#define IsOut(a)  ((int)((a)&F4_WRITE))
#define IsPipe(a) ((a)&F4_PIPE)

#ifndef UNIX                           /* MS-OSes have no pipes */
#define NOPIPES
#endif

#ifdef __DJGPP__                       /* can't get DJGPP to work */
#define NOPIPES
#endif
                                       /* define 4tH I/O device */
typedef struct {
  int     Channel;                     /* currently used channel */
  FILE*   Device;                      /* external device opened */
  FILE* (*Connect)    (const char*, const char*);
  int   (*Disconnect) (FILE *);
  FILE   *StdDev;                      /* standard device */
} DEV4TH;
                                       /* 4tH I/O devices */
static DEV4TH Stream  [MAXDEVS];
static FILE  *Channel [2] = {NULL, NULL};

static unsigned ToS, ToRS;             /* top of stack/top of ret. stack */
static unsigned Hld;                   /* equivalent of forth HLD */
static unsigned CharSiz;               /* size of Character Segment */
static unsigned Handler;               /* throw handler */
static unit    *CharSeg;               /* points to string Character Segment */
static cell    *Stack;                 /* pointer to stack */
static cell    *Base;                  /* radix status */
static Hcode   *Applet;                /* externalize Object */
static jmp_buf  Thrown;                /* buffer for throw error-return */
                                       /* THROW prototype */
static void throw __PROTO((cell ErrCode));

#ifdef NOPIPES                         /* dummy pipe functions */
static FILE* Popen(const char* a, const char* b) { if (a==b); return (NULL); }
static int Pclose(FILE *a) { if (a); return (0); }
#else                                  /* more reasonable pclose */
#ifndef ARCHAIC
   static int Pclose (FILE *p)
#else
   static int Pclose (p) FILE *p;
#endif                                 /* no error if child died */
{
  return ((pclose (p)) && (errno != ECHILD) ? -1 : 0);
}

#ifndef ARCHAIC                        /* relatively safe popen() */
  static FILE *Popen (const char *f, const char *m)
#else
  static FILE *Popen (f, m) const char *f; const char *m;
#endif
{
  FILE *p;                             /* filepointer to pipe */
  long  x;                             /* simple counter */
  int   s = 0;                         /* status of child */

  p = popen (f, m);                    /* perform a normal popen() */
  for (x = 0; x < PIPEWAIT && s == 0; x++) waitpid (-1, &s, WNOHANG);
  return (s ? NULL : p);               /* check the status for a while */
}                                      /* return NULL if not normal */
#endif


#ifndef ARCHAIC
  static void push (cell val)          /* adds a value to the stack */
#else
  static void push (val) cell val;
#endif

{
  if (ToS < ToRS) Stack [ToS++] = val;
  else throw (M4SOVFLW);
}


#ifndef ARCHAIC
  static cell pop (void)               /* pops a value from the stack */
#else
  static cell pop ()
#endif

{
  if (ToS > S0) return (Stack [--ToS]);
  else
    {
      throw (M4SEMPTY);
      return (CELL_MIN);
    }
}


#ifndef ARCHAIC
  static void rpush (cell val)         /* adds a value on the return stack */
#else
  static void rpush (val) cell val;
#endif

{
  if (ToRS > ToS) Stack [ToRS--] = val;
  else throw (M4ROVFLW);
}


#ifndef ARCHAIC
  static cell rpop (void)              /* pops a value from the return stack */
#else
  static cell rpop ()
#endif

{
  if (ToRS < R0) return (Stack [++ToRS]);
  else
    {
      throw (M4REMPTY);
      return (CELL_MIN);
    }
}


#ifndef ARCHAIC
  static void branch (cell address)    /* sets program counter to addr */
#else
  static void branch (address) cell address;
#endif

{
  if ((address < (cell) Applet->CodeSiz) && (address > -2L))
     Applet->ErrLine = (int) address;
  else throw (M4BADTOK);
}


/*
This is the equivalent of Forths THROW. Since it still needs an interface
to the stacks, the danger of cascaded THROWs exists. This is prevented by
the boolean PassThru, which is set when popping starts. Subsequent calls to
throw() simply "pass through".
*/

#ifndef ARCHAIC
  static void throw (cell ErrCode)
#else
  static void throw (ErrCode) cell ErrCode;
#endif

{
  static char PassThru = FALSE;        /* to prevent cascaded THROWs */
  cell StackPointer;                   /* temporary storage for SP */
  cell ReturnAddress;                  /* temporary storage for return addr */
                                       /* THROW 0 does nothing */
  if ((ErrCode) && (! PassThru))
     {
       if (Handler)                    /* if a previous handler */
          {                            /* handler contains RP */
            if ((Handler > R0) || (Handler < (cell) ToRS))
              ErrCode = M4BADPTR;      /* Return Stack may only shrink */
            else                       /* else handler is invalid */
              {                        /* restore RP from handler */
                PassThru = TRUE;       /* danger of cascading THROWs */
                ToRS = Handler;        /* retrieve handler from Return Stack */
                ReturnAddress = rpop(); /* retrieve return address */
                Handler = (unsigned) rpop();
                StackPointer = rpop(); /* retrieve SP from Return Stack */
                                       /* SP might have shrunk after CATCH */
                if ((StackPointer < S0) || (StackPointer > (cell) ToRS))
                  {                    /* SP may rise to previous position */
                    ErrCode = M4BADPTR; /* SP may not violate Return Stack */
                    PassThru = FALSE;  /* THROWs are welcome again */
                  }
                else                   /* restore SP */
                  {                    /* drop item from SP; add ErrCode */
                    ToS = (unsigned) StackPointer;
                    pop (); push ((cell) -ErrCode);
                    branch (ReturnAddress + 2);
                    PassThru = FALSE;  /* THROWs are welcome again */
                    longjmp (Thrown, EXCEPTION);
                  }                    /* jump to main loop */

              }
          }                            /* no or invalid handler */
       if ((ErrCode < 0) || (ErrCode > LastMsg4th)) Applet->ErrNo = M4THROW;
       else Applet->ErrNo = (unsigned) ErrCode;
       longjmp (Thrown, EXCEPTION);    /* jump to main loop */
     }
}


/* gets a value from the String User Area */

#ifndef ARCHAIC
  static unit fetch (cell address)
#else
  static unit fetch (address) cell address;
#endif

{
  if ((address >= TIB) && (address < CharSiz))
     return (CharSeg [(unsigned) address]);
  else
     throw (M4BADADR);
  return (' ');
}


/* stores a value in the String User Area */

#ifndef ARCHAIC
  static void store (cell address, unit value)
#else
  static void store (address, value) cell address; unit value;
#endif

{
  if ((address >= TIB) && (address < CharSiz))
     CharSeg [(unsigned) address] = value;
  else
     throw (M4BADADR);
}


/*
This function is roughly equivalent with Forths <#
*/

#ifndef ARCHAIC
  static void lsharp (void)
#else
  static void lsharp ()
#endif

{
  Hld = (PAD + PADSIZ); store ((cell) --Hld, '\0');
}


/*
This function is equivalent with Forths SIGN
*/

#ifndef ARCHAIC
  static void sign (cell x)
#else
  static void sign (x) cell x;
#endif

{
  if (x < 0) store ((cell) --Hld, '-');
}


/*
This function is roughly equivalent with Forths #
*/

#ifndef ARCHAIC
  static unit sharp (cell *x)
#else
  static unit sharp (x) cell *x;
#endif

{
  unit y;

  if ((*Base < 2) || (*Base > 36))
     {
       throw (M4BADRDX);
       return (' ');
     }

  *x = pop ();
  y = (unit) (*x % *Base);
  push (*x /= *Base);
  return (y += (y > 9) ? '7' : '0');
}


/*
This function is equivalent of Forths #S
*/

#ifndef ARCHAIC
  static void sharpS (cell *x)
#else
  static void sharpS (x) cell *x;
#endif

{
  do store ((cell) --Hld, sharp (x));
  while (*x);
}


/*
This baby simply prints a single character. Equivalent of Forths EMIT
*/

#ifndef ARCHAIC
  static void emit (unit c)
#else
  static void emit (c) unit c;
#endif

{
  if (! Channel [COUT]) throw (M4IOERR);
  else if ((fputc ((int) c, Channel [COUT])) == EOF) throw (M4IOERR);
}


/*
This function prints y characters from address x. Equivalent of Forths TYPE.
*/

#ifndef ARCHAIC
  static void type (cell x, cell y)
#else
  static void type (x, y) cell x; cell y;
#endif

{
  for (;y; y--, x++) emit (fetch (x));
}


/*
Prints a number according to the current radix. Roughly equivalent to D.R
*/

#ifndef ARCHAIC
  static void dotR (cell *p, cell q)
#else
  static void dotR (p, q) cell *p; cell q;
#endif

{
  cell t;

  *p = pop (); t = *p;                          /* t holds the sign */
  if (*p < 0) *p = -(*p); push (*p);            /* make number positive */
  lsharp (); sharpS (p); sign (t); *p = pop (); /* <# #S SIGN #> */
  for (;strlen ((char*) CharSeg + Hld) < q;)
      store ((cell) --Hld, ' ');
  type ((cell) Hld, (cell) (PAD + PADSIZ) - (Hld + 1));
}


/*
This routine reads J characters from input device DEV and places them at
address I. It returns the number of characters actually read.
Leaves ASCIIZ string when reading a textdevice.
*/

#ifndef ARCHAIC
  static cell accept (cell i, cell j, char mode)
#else
  static cell accept (i, j, mode) cell i; cell j; char mode;
#endif

{
  int  c;                              /* char read */
  char OK = TRUE;                      /* OK to read next char */
  cell t = 0;                          /* counter */
                                       /* check for device */
  if (! Channel [CIN]) throw (M4IOERR);
  else
    {                                  /* device is keyboard: textmode */
      if (Channel [CIN] == stdin) mode = 't';

      do {
          c = fgetc (Channel [CIN]);   /* get character */
          if ((c == EOF) || ((mode == 't') && (c == (int) '\n'))) OK = FALSE;
          else if ((mode == 'b') ||    /* stop on EOF or <lf> in text mode */
            ((c != (int) '\r') && (c != EOTu) && (c != EOTm)))
            {                          /* store the character and incr count */
              store (i + t, (unit) c); /* when in binary mode or when in */
              t++;                     /* textmode and no <lf> or EOF */
            }                          /* has been read (these are skipped) */
      } while ((OK) && (t < j));       /* until count reached or EOF */
                                       /* when reading from keyboard */
      if ((OK) && (Channel [CIN] == stdin))
         do { c = fgetc (Channel [CIN]); } while ((c != EOF) && (c != '\n'));
                                       /* empty buffer (read till <lf>) */
      if (mode == 't')
         {
           store (i + t, '\0');        /* when in text mode terminate string */
           if ((t == 0) && (c == EOF)) return (-1);
         }                             /* return -1 when nothing read */
    }

  return (t);                          /* return #chars read */
}


/*
This routine closes an input- or an output file. ErrCode determines which
error is thrown. M4NOERR will bypass the throw() routine and force the close.
*/

#ifndef ARCHAIC
  static void CloseStream (int StrNo, unsigned ErrCode)
#else
  static void CloseStream (StrNo, ErrCode) int StrNo; unsigned ErrCode;
#endif

{                                      /* is it a valid stream */
  if ((StrNo < (cell) CIN) || (StrNo >= MAXDEVS)) throw (M4BADDEV);
  else                                 /* ignore unopened streams, CIN, COUT */
    if ((Stream [StrNo].Device != NULL) && (StrNo > COUT))
      {                                /* redirect channel, no matter what */
        Channel [Stream [StrNo].Channel] = Stream [StrNo].StdDev;
        if (Stream [StrNo].Disconnect (Stream [StrNo].Device))
          throw (ErrCode);             /* throw on close error */
        Stream [StrNo].Device = NULL;  /* signal external device closed */
      }
}


/*
This routine opens an input- or an output file.
*/

#ifndef ARCHAIC
  static cell OpenStream (int IsWrite, cell Pipe, char *Name, char *Mode)
#else
  static cell OpenStream (IsWrite, Pipe, Name, Mode) int IsWrite; cell Pipe;
                          char *Name; char *Mode;
#endif

{
#ifndef WEB4TH
  int x;

  for (x = COUT + 1; x < MAXDEVS; x++) /* do not search CIN and COUT */
      if (Stream [x].Device == NULL)   /* find unopened stream */
         {                             /* initialize stream */
           Stream [x].Channel = IsWrite;
           Stream [x].Disconnect = (Pipe ? Pclose : fclose);
           Stream [x].Connect =    (Pipe ? Popen  : fopen);
           Stream [x].StdDev =  (IsWrite ? stdout : stdin);
           return (((Stream [x].Device = Stream [x].Connect
                    (Name, Mode)) == NULL ? F_F : (cell) x ));
         }                             /* open stream */
#endif
  return (F_F);

}


/*
This routine simply copies the string constant 'constr' to the location
in the Character Segment, identified by 'str'. Used by e.g. COPY and ARGS.
*/

#ifndef ARCHAIC
  static void toCharSeg (cell str, char* constr)
#else
  static void toCharSeg (str, constr) cell str; char* constr;
#endif

{
  for (;*constr; constr++, str++) store (str, (unit) *constr);
  store (str, '\0');
}


/*
Main routine and the only external.
*/

#ifndef ARCHAIC
  cell exec_4th (Hcode *Object, unsigned ArgN, char **ArgS,
                 unsigned VarN, ...)
#else
  cell exec_4th (Object, ArgN, ArgS, VarN) Hcode *Object;
                 unsigned ArgN; char **ArgS; unsigned VarN;
#endif

{
  unsigned VarNum;                     /* calculate number of variables */
  unsigned VarOffset;                  /* offset of user variables */
  unsigned x;                          /* temporary counter */
  cell a, b, c;                        /* internal stack variables */
  time_t Now;                          /* temporary time variable */
  unit  t;                             /* temporary char */
  char *p;                             /* temporary pointer */
  cell *Vars;                          /* pointer to Variable Area */
  cell *In;                            /* equivalent of forth >IN */
  cell *Result;                        /* return value for apps */
  va_list Vals;                        /* list of preloaded variables */

  if (! Object) return (CELL_MIN);   /* check if object is reliable */
  else if (! Object->Reliable) return (CELL_MIN);

  Applet = Object;                     /* externalize argument */
  Object->ErrNo = M4NOERRS;            /* reset errorcode */
  Object->ErrLine = 0;                 /* reset linepointer */

  VarOffset = VarN + VAR4TH;           /* reserve space for BASE, >IN */
  VarNum = Object->Variables + VarOffset;

  if ((Stack = (cell*) calloc (STACKSIZ + VarNum, sizeof (cell))) == NULL)
     {                                 /* allocate Integer Segment */
       Object->ErrNo = M4NOMEM;
       return (CELL_MIN);
     }

  Vars = Stack + STACKSIZ;             /* set pointer to Variable Area */
  ToRS = R0;                           /* reset return stack pointer */
  ToS  = S0;                           /* reset stackpointer */
  Handler = 0;                         /* initialize HANDLER */

  CharSiz = Object->Strings + TIBSIZ + PADSIZ;

  if ((CharSeg = (unit*) calloc (CharSiz + 1, sizeof (unit))) == NULL)
     {                                 /* allocate Character Segment */
       free (Stack);
       Object->ErrNo = M4NOMEM;
       return (CELL_MIN);
     }

  Base     = &(Vars [0]);              /* assign pointer to BASE */
  In       = &(Vars [1]);              /* assign pointer to >IN */
  Result   = &(Vars [2]);              /* assign pointer to RESULT */

  *Base    = 10;                       /* initialize BASE */
  *In      = TIB;                      /* initialize >IN */
  *Result  = CELL_MIN;                 /* initialize RESULT */
                                       /* preload variables */
  va_start (Vals, VarN);
  for (x = VAR4TH; x < VarOffset; x++) Vars [x] = va_arg (Vals, cell);
  va_end (Vals);

  srand (SEED);                        /* initialize randomizer */
                                       /* initialize streams */
  for (x = CIN; x < MAXDEVS; x++) Stream [x].Device = NULL;
  Channel [CIN]  = Stream [CIN].Device  = Stream [CIN].StdDev  = stdin;
  Channel [COUT] = Stream [COUT].Device = Stream [COUT].StdDev = stdout;
  Stream [CIN].Channel = CIN; Stream [COUT].Channel = COUT;
                                       /* read and write from tty */
/*
This is the main calculation engine. Most operators can be reduced to pop() and
push() calls. It executes until there are no more operators left or QUIT is
called.
*/

  Object->ErrLine = 0;                 /* setup initial entry point */
  (void) setjmp (Thrown);              /* Return from error */

  if ((! Object->ErrNo) && (Object->ErrLine >= 0))
    for (;(Object->ErrLine < Object->CodeSiz); Object->ErrLine++)
      switch (Object->CodeSeg [Object->ErrLine].Word)
      {
        case (NOOP):    break;
        case (LITERAL): push (Object->CodeSeg [Object->ErrLine].Value); break;
        case (BRANCH):  branch (Object->CodeSeg [Object->ErrLine].Value);
                        break;
        case (BRANCH0): if (pop () == F_F)
                           branch (Object->CodeSeg [Object->ErrLine].Value);
                        break;
        case (LOOP):    a = rpop (); b = rpop (); a++;
                        if (a < b)
                           {
                             rpush (b); rpush (a);
                             branch (Object->CodeSeg [Object->ErrLine].Value);
                           }
                        break;
        case (DO):      a = pop (); rpush (pop ()); rpush (a); break;
        case (CALL):    rpush ((cell) Object->ErrLine);
                        branch (Object->CodeSeg [Object->ErrLine].Value);
                        break;
        case (EXIT):    branch (rpop ()); break;
        case (DUP):     a = pop (); push (a); push (a); break;
        case (DROP):    (void) pop (); break;
        case (SWAP):    a = pop (); b = pop (); push (a); push (b); break;
        case (ROT):     a = pop (); b = pop (); c = pop (); push (b); push (a);
                        push (c); break;
        case (OVER):    b = pop (); a = pop (); push (a); push (b); push (a);
                        break;
        case (VARIABLE):push ((cell) VarOffset +
                             Object->CodeSeg [Object->ErrLine].Value);
                        break;
        case (VALUE):   push (Vars [VarOffset + (unsigned)
                             Object->CodeSeg [Object->ErrLine].Value]);
                        break;
        case (RCOPY):   a = rpop (); rpush (a); push (a); break;
        case (STORE):   a = pop ();
                        if ((a < (cell) VarNum) && (a >= 0L))
                           Vars [(unsigned) a] = pop ();
                        else throw (M4BADVAR); break;
        case (FETCH):   a = pop ();
                        if ((a < (cell) VarNum) && (a >= 0L))
                           push (Vars [(unsigned) a]);
                        else throw (M4BADVAR); break;
        case (ADDSTORE):a = pop ();
                        if ((a < (cell) VarNum) && (a >= 0L))
                           Vars [(unsigned) a] += pop ();
                        else throw (M4BADVAR); break;
        case (QDO):     a = pop (); b = pop ();
                        if (a == b)
                           branch (Object->CodeSeg [Object->ErrLine].Value);
                        else
                           {
                             rpush (b); rpush (a);
                           }
                        break;
        case (PLOOP):   a = rpop (); b = rpop (); c = pop (); a += c;
                        if ((c > 0L) ? (a < b) : (a > b))
                           {
                             rpush (b); rpush (a);
                             branch (Object->CodeSeg [Object->ErrLine].Value);
                           }
                        break;
        case (LEAVE):   a = rpop (); (void) rpop (); rpush (a); rpush (a);
                        break;
        case (EQ0):     push (pop() == 0L ? F_T : F_F); break;
        case (NE0):     push (pop() != 0L ? F_T : F_F); break;
        case (PLUS):    push (pop () + pop ()); break;
        case (EQ):      push (pop () == pop () ? F_T : F_F); break;
        case (MOD):     b = pop ();
                        if (b != 0L) push (pop () % b);
                        else throw (M4DIVBY0);
                        break;
        case (MINUS):   b = pop (); push (pop () - b); break;
        case (INC):     a = pop (); push (++a); break;
        case (DEC):     a = pop (); push (--a); break;
        case (MUL):     push (pop () * pop ()); break;
        case (DIV):     b = pop ();
                        if (b != 0L) push (pop () / b);
                        else throw (M4DIVBY0);
                        break;
        case (RIGHT2):  push (pop() >> 1L); break;
        case (SLASHMOD):b = pop (); a = pop ();
                        if (b != 0L)
                           {
                             push (a % b); push (a / b);
                           }
                        else throw (M4DIVBY0);
                        break;
        case (GT):      b = pop (); push (pop () > b ? F_T : F_F); break;
        case (MIN):     b = pop (); a = pop (); push ((a < b ? a : b)); break;
        case (MAX):     b = pop (); a = pop (); push ((a > b ? a : b)); break;
        case (STARSLSH):a = pop ();
                        if (a != 0L) push ((pop () * pop ()) / a);
                        else throw (M4DIVBY0);
                        break;
        case (LEFT2):   push (pop () << 1L); break;
        case (STSLHMOD):a = pop (); b = pop (); c = pop ();
                        if (a != 0L)
                           {
                             push ((b * c) % a); push ((b * c) / a);
                           }
                        else throw (M4DIVBY0);
                        break;
        case (ABS):     a = pop (); push ((a < 0L ? -a : a)); break;
        case (NEGATE):  push (-(pop ())); break;
        case (OR):      b = pop (); push (pop () | b); break;
        case (AND):     b = pop (); push (pop () & b); break;
        case (XOR):     b = pop (); push (pop () ^ b); break;
        case (SHIFT):   b = pop (); a = pop ();
                        push (b < 0 ? a >> -(b) : a << b);
                        break;
        case (INVERT):  push (~(pop ())); break;
        case (LT):      b = pop (); push (pop () < b ? F_T : F_F); break;
        case (GE):      b = pop (); push (pop () >= b ? F_T : F_F); break;
        case (LE):      b = pop (); push (pop () <= b ? F_T : F_F); break;
        case (NE):      push (pop () != pop () ? F_T : F_F); break;
        case (LT0):     push (pop () < 0L ? F_T : F_F); break;
        case (GT0):     push (pop () > 0L ? F_T : F_F); break;
        case (RGET):    push (rpop ()); break;
        case (RPUT):    rpush (pop ()); break;
        case (RCOPY3):  a = rpop (); b = rpop (); c = rpop (); rpush (c);
                        rpush (b); rpush (a); push (c); break;
        case (PRINT):   for (p = Object->StringSeg + (unsigned)
                             Object->CodeSeg [Object->ErrLine].Value;
                             *p; p++) emit ((unit) *p);
                        break;
        case (CR):      emit ('\n'); break;
        case (SPACES):  a = pop(); for (b = 0L; b < a; b++) emit (' '); break;
        case (DOT):     dotR (&a, 0); emit (' '); break;
        case (DOTR):    dotR (&b, pop ()); break;
        case (EMIT):    emit ((unit) pop ()); break;
        case (WAIT):    a = pop ();
#ifndef _Windows
                        sleep ((unsigned) a);
#endif
                        break;
        case (OCTAL):   *Base = 8; break;
        case (DECIMAL): *Base = 10; break;
        case (HEX):     *Base = 16; break;
        case (QUIT):    Object->ErrLine = Object->CodeSiz - 1; break;
        case (TO):      Vars [VarOffset + (unsigned)
                             Object->CodeSeg [Object->ErrLine].Value] = pop ();
                        break;
        case (EXECUTE): rpush ((cell) Object->ErrLine); branch (pop ()); break;
        case (FETCHPRM):a = pop ();
                        if ((a >= 0L) && (a < (cell) Object->CodeSiz))
                           push (Object->CodeSeg [(int) a].Value);
                        else throw (M4BADTOK); break;
        case (TIME):    push (time (&Now)); break;
        case (CLSPAREN):if (! pop()) throw (M4ASSERT); break;
        case (RANDOM):  push ((cell) (rand () & RNDMASK)); break;
        case (CFETCH):  push ((cell) fetch (pop ())); break;
        case (CSTORE):  a = pop (); store (a, (unit) pop ()); break;
        case (LSHARP):  lsharp (); break;
        case (RSHARP):  (void) pop (); push ((cell) Hld);
                        push ((cell) (PAD + PADSIZ) - (Hld + 1)); break;
        case (HOLD):    store ((cell) --Hld, (unit) pop ()); break;
        case (SIGN):    a = pop (); sign (pop ()); push (a); break;
        case (SHARP):   store ((cell) --Hld, sharp (&a)); break;
        case (SHARPS):  sharpS (&a); break;
        case (TYPE):    a = pop (); type (pop (), a); break;
        case (FILL):    t = (unit) pop (); b = pop (); c = pop ();
                        for (;b; b--, c++) store (c, t);
                        break;
        case (COUNT):   a = pop ();
                        if ((a >= TIB) && (a < CharSiz))
                           {
                             push (a);
                             push ((cell) strlen ((char*) CharSeg +
                                  (unsigned) a));
                           }
                        else throw (M4BADADR);
                        break;
        case (NUMBER):  if ((*Base < 2) || (*Base > 36))
                           throw (M4BADRDX);
                        else
                          {
                            a = pop ();
                            if ((a >= TIB) && (a < CharSiz))
                              {
                                b = strtoc  ((char*)  CharSeg + (unsigned) a,
                                    &p, (int) *Base);
                                if  ((p ==  ((char*) (CharSeg + (unsigned) a) +
                                     strlen ((char*)  CharSeg + (unsigned) a)))
                                && (strlen ((char*)  CharSeg + (unsigned) a)
                                     > 0)) push (b);
                                else push (CELL_MIN);
                              }
                            else throw (M4BADADR);
                          }
                        break;
        case (SKIP):    a = pop ();
                        do
                          t = fetch ((*In)++);
                        while ((t == (unit) a) && (t));
                        --(*In); break;
        case (TOKEN):   a = pop (); b = PAD;
                        do
                          store (b++, (t = fetch ((*In)++)));
                        while ((t != (unit) a) && (t));
                        store (--b, '\0');
                        if (! t) (*In)--;
                        push ((cell) PAD); break;
        case (COPY):    a = pop (); b = pop ();
                        if ((b < 0) || (b >= Object->StringSiz))
                           throw (M4BADSTR);
                        else
                           {
                             toCharSeg (a, Object->StringSeg + (unsigned) b);
                             push (a);
                           }
                        break;
        case (CMOVE):   a = pop (); b = pop (); c = pop ();
                        if (c < b)
                           {
                             b += (a - 1); c += (a - 1);
                           }
                        for (x = (unsigned) a; x;
                            (c < b) ? (x--, b--, c--) : (x--, b++, c++)
                        ) store (b, fetch (c));
                        break;
        case (TRAILING):a = pop (); b = pop ();
                        do a--;
                        while (fetch (b + a) == ' ');
                        push (b); push (++a);
                        break;
        case (ACCEPT):  a = pop ();
                        push (accept (pop (), a, 'b'));
                        break;
        case (REFILL):  push (accept (TIB, TIBSIZ - 1, 't') == -1 ? F_F : F_T);
                        *In = TIB; break;
        case (THROW):   throw (-(pop())); break;
        case (CATCH):   rpush ((cell) ToS); rpush ((cell) Handler);
                        rpush ((cell) Object->ErrLine); Handler = ToRS;
                        branch (pop ()); break;
        case (CAUGHT):  Handler = (unsigned) rpop (); (void) rpop();
                        push (0L); break;
        case (LAST):    push ((cell) VarNum - 1); break;
        case (FIRST):   push ((cell) VarOffset); break;
        case (HI):      push ((cell) CharSiz - 1); break;
        case (SPFETCH): push ((cell) ToS); break;
        case (RPFETCH): push ((cell) ToRS); break;
        case (ARGN):    push ((cell) ArgN); break;
        case (SQUOTE):  toCharSeg (PAD, Object->StringSeg + (unsigned)
                        Object->CodeSeg [Object->ErrLine].Value);
                        push ((cell) PAD);
                        push ((cell) strlen ((char*) CharSeg + PAD));
                        break;
        case (ARGS):    a = pop ();
                        if ((a < 0) || (a >= ArgN)) throw (M4BADSTR);
                        else
                           {
                             toCharSeg (PAD, ArgS [(unsigned) a]);
                             push ((cell) PAD);
                           }
                        break;
        case (USE):     a = pop ();
                        if ((a < (cell) CIN) || (a >= MAXDEVS))
                          throw (M4BADDEV);
                        else if (Stream [(int)a].Device != NULL)
                          Channel [Stream [(int)a].Channel] =
                                   Stream [(int)a].Device;
                        break;
        case (FCLOSE):  CloseStream (pop (), M4IOERR); break;
        case (FOPEN):   a = pop (); (void) pop (); b = pop ();
                        if ((b < TIB) || (b >= CharSiz)) throw (M4BADADR);
                        if (IsPipe (a)) a &= (~(F4_APPND + F4_TEXT));

                        switch ((int) a)
                        {
                          case (F4_WRITE):                  p = "wb"; break;
                          case (F4_WRITE+F4_TEXT):          p = "wt"; break;
                          case (F4_WRITE+F4_APPND):         p = "ab"; break;
                          case (F4_WRITE+F4_APPND+F4_TEXT): p = "at"; break;
                          case (F4_PIPE):                   p = "r" ; break;
                          case (F4_WRITE+F4_PIPE):          p = "w" ; break;
                          default:                          p = "rb"; break;
                        }

                        push (OpenStream (IsOut (a), IsPipe (a),
                             (char*) CharSeg + (unsigned) b, p));
                        break;
        default:        throw (M4BADOBJ); break;
      }
                                       /* close files forced */
   for (a = (cell) (COUT + 1); a < MAXDEVS; a++)
        CloseStream ((int) a,  M4NOERRS);

   a = *Result;                        /* get result */

   free (CharSeg);                     /* free environment */
   free (Stack);

   if (Object->ErrNo) return (CELL_MIN);
   else return (a);                    /* return value */
}











