/*
+-------+-----------------+------+-----------------+-------+-----------------+
|Program| HGEN_4TH.C      |Header| 4TH.H           |Library| CEXT            |
+-------+-----------------+------+-----------------+-------+----------+------+
|HISTORY                                           |    Programmer    | Date |
+--------------------------------------------------+------------------+------+
| Turbo C V2.0 version                             | HanSoft          |311099|
| Linux V2.2.10 version                            | HanSoft          |210600|
+--------------------------------------------------+------------------+------+
|MODIFICATION LOG                                  |    Programmer    | Date |
+--------------------------------------------------+------------------+------+
| Changed 'bytecode' to 'unit' in text             | HanSoft          |021199|
| Changed source FP to sourcename (check if open)  | HanSoft          |041199|
| Fixed major bug; WriteError properly initialized | HanSoft          |111199|
| Fixed small error in K&R prototype               | HanSoft          |291299|
| Fixed bug: sourcefile not closed on error        | HanSoft          |210600|
+--------------------------------------------------+------------------+------+
|REMARKS                                           |    Programmer    | Date |
+--------------------------------------------------+------------------+------+
| None                                             |                  |      |
+--------------------------------------------------+------------------+------+

  Copyright (C) 1997,9 J.L. Bezemer

  This file is part of 4tH

  4tH is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

4tH function to write a C include file, which contains the embedded HX code
of an HX file in a named variable.

Prototype: char hgen_4th (char *HxFile, char *SourceFile, char *VarName)
Returns  : TRUE if an error occured, FALSE if successful
Memory   : None
Example  :


#include "4th.h"
#include <stdlib.h>

#ifndef ARCHAIC
  int main (int argc, char **argv)
#else
  int main (argc, argv) int argc; char **argv;
#endif

{
  if (argc == 3)
    {
      if (! hgen_4th (argv [1], argv [2], "EmbeddedHX"))
         return (EXIT_SUCCESS);
    }
  return (EXIT_FAILURE);
}

Input    : None
Output   : depends on program
Related  : dump_4th(), cgen_4th()
*/

#ifdef USRLIB4TH
#include <4th.h>
#include <sys/cmds_4th.h>
#else
#include "4th.h"
#include "cmds_4th.h"
#endif

#include <stdarg.h>
                                       /* HX file headers */
#define HeadHx  "static unit %s [] = {\n  "
#define TailHx  "\'\\x%02x\'\n};\n"
                                       /* TRUE if a write error occured */
static char WriteError;


/*
This function does the actual writing of the C code to the device. If an
error occurs, it sets WriteError, which prevents further writing.
*/

#ifndef ARCHAIC
  static void WriteSource (FILE *SourceFp, char *Format, ...)
#else
  static void WriteSource (SourceFp, Format) FILE *SourceFp; char* Format;
#endif

{
  va_list  (args);
  va_start (args, Format);

  if (! WriteError)
     if (vfprintf (SourceFp, Format, args) == 0) WriteError = TRUE;

  va_end (args);
}


/*
This routine prints a single line of unsigned bytes to the output device.
*/

#ifndef ARCHAIC
  static void WriteHxLine (FILE* OutDev, unit *Buffer, size_t MaxLen)
#else
  static void WriteHxLine (OutDev, Buffer, MaxLen) FILE* OutDev; unit *Buffer;
                           size_t MaxLen;
#endif

{
  size_t x;                            /* simple counter */

  for (x = 0; x < MaxLen; x++)         /* print all bytes */
      WriteSource (OutDev, "\'\\x%02x\', ", (int) Buffer [x]);
}


/*
This function does the actual creation of the C source. It returns TRUE
if an error occured during the actual writing.
*/

#ifndef ARCHAIC
  static void GenerateSource (FILE *HXp, FILE *SourceFp, char *VarName)
#else
  static void GenerateSource (HXp, SourceFp, VarName) FILE *HXp;
                              FILE *SourceFp; char *VarName;
#endif

{
  size_t BufFill;                      /* number of bytes read */
  unit   Buf [UNITSLINE];              /* read buffer */

  WriteSource (SourceFp, HeadHx, VarName);
                                       /* print all full lines */
  while ((BufFill = fread (Buf, sizeof (char), UNITSLINE, HXp))
         == UNITSLINE)
    {
       WriteHxLine (SourceFp, Buf, UNITSLINE);
       fprintf (SourceFp, "\n  ");     /* print terminating EOL */
    }
                                       /* print last line and footer */
  if (BufFill == 0)
     WriteSource (SourceFp, TailHx, 0);
  else
     {
       WriteHxLine (SourceFp, Buf, BufFill - 1);
       WriteSource (SourceFp, TailHx, Buf [BufFill - 1]);
     }
}


/*
This function creates C source as an include file from a named HX file. Note
it doesn't check whether the file is a valid HX file.
*/

#ifndef ARCHAIC
  char hgen_4th (char *HxFile, char *SourceFile, char *VarName)
#else
  char hgen_4th (HxFile, SourceFile, VarName) char *HxFile;
                 char *SourceFile; char *VarName;
#endif

{
  FILE *HXp;                           /* HX file */
  FILE *SourceFp;                      /* C source file */

  WriteError = FALSE;                  /* reset error flag */

  if ((HXp = fopen (HxFile, "rb")) != NULL);
     if ((SourceFp = fopen (SourceFile, "wt")) == NULL) fclose (HXp);
     else                              /* if opening failed, close HX file */
        {                              /* generate the source */
          GenerateSource (HXp, SourceFp, VarName);
          if (fclose (SourceFp)) WriteError = TRUE;
          if (! (fclose (HXp))) return (WriteError);
        }                              /* close all files and return status */
                                       /* return true on all errors */
  return (TRUE);
}
