/*
+-------+-----------------+------+-----------------+-------+-----------------+
|Program| LOAD_4TH.C      |Header| 4TH.H           |Library| CEXT            |
+-------+-----------------+------+-----------------+-------+----------+------+
|HISTORY                                           |    Programmer    | Date |
+--------------------------------------------------+------------------+------+
| Turbo C V2.0 version                             | HanSoft          |250495|
+--------------------------------------------------+------------------+------+
|MODIFICATION LOG                                  |    Programmer    | Date |
+--------------------------------------------------+------------------+------+
| Fixed string-area bug                            | HanSoft          |280495|
| Definitely fixed void bug                        | HanSoft          |010595|
| Added ErrLin4th                                  | HanSoft          |030595|
| New Version4th added                             | HanSoft          |140995|
| Updated errors                                   | HanSoft          |220995|
| Added support for Hcode                          | HanSoft          |120296|
| Changed header to check alignment; fixed bug     | HanSoft          |130396|
| Removed LoadBlock(); added LoadByte(), LoadCell()| HanSoft          |100496|
| Added LoadInfo(), LoadHeader() and LoadCode()    | HanSoft          |100496|
| Added LoadString(), LoadCheck()                  | HanSoft          |100496|
| Fixed bugs in LoadString() and LoadCheck()       | HanSoft          |110496|
| Added RaiseError()                               | HanSoft          |120496|
| Rewrote most for new Hcode format                | HanSoft          |021096|
| Renamed Appl4th to App4th                        | HanSoft          |031296|
| Added LOCAL_H support                            | HanSoft          |220597|
| Removed <limits.h>                               | HanSoft          |250797|
| Added BRANCHXT; changed header fieldname         | HanSoft          |120997|
| Removed BRANCHXT                                 | HanSoft          |170997|
| Added QDO                                        | HanSoft          |151097|
| Forgot to close file if calloc() failed          | HanSoft          |171097|
| Removed EasyC syntax                             | HanSoft          |150198|
| Updated version info                             | HanSoft          |180298|
| Added SQUOTE                                     | HanSoft          |200298|
| Added inst_4th() and support functions           | HanSoft          |081298|
| Added some documentation                         | HanSoft          |101298|
| Added LoadValue(); more compact HX format        | HanSoft          |140999|
| Cleaned up encoding constants                    | HanSoft          |150999|
| Replaced bytecode datatype by unit datatype      | HanSoft          |301099|
| Return CodeSiz in ErrLine when successful        | HanSoft          |031199|
| Fixed major bug; CSIn now properly initialized   | HanSoft          |111199|
+--------------------------------------------------+------------------+------+
|REMARKS                                           |    Programmer    | Date |
+--------------------------------------------------+------------------+------+
| None                                             |                  |      |
+--------------------------------------------------+------------------+------+

  Copyright (C) 1997,9 J.L. Bezemer

  This file is part of 4tH

  4tH is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


These functions load HX bytecode into memory. Compatibility is tested!

Prototype: Hcode* load_4th (char *filename)
Returns  : Object is compatible and loadable; NULL if not.
Memory   : header, object and string-area's are allocated on success
Example  :

#include "4th.h"
#include <stdlib.h>

#ifndef ARCHAIC
  int main (int argc, char **argv)
#else
  int main (argc, argv) int argc; char **argv;
#endif

{
  Hcode *object;

  object = load_4th ("easter.hx");
  dump_4th (object, stdout, 0, -1);
  free_4th (object);
  return (EXIT_SUCCESS);
}

Input    : None
Output   : depends on program
Related  : inst_4th()


Prototype: Hcode* inst_4th (unit *embedded)
Returns  : Object is compatible and loadable; NULL if not.
Memory   : header, object and string-area's are allocated on success
Example  :

#include "4th.h"
#include <stdlib.h>

#ifndef ARCHAIC
  int main (int argc, char **argv)
#else
  int main (argc, argv) int argc; char **argv;
#endif

{
  static unit EmbeddedHX [] = {
    '\x01', '\x02', '\x04', '\x00', '\xff', '\xff', '\xff', '\x7f',
    '\x04', '\x3b', '\x03', '\x08', '\x02', '\x02', '\x02', '\x0d',
    '\x08', '\x08', '\x05', '\x08', '\x02', '\x48', '\x65', '\x6c',
    '\x6c', '\x6f', '\x20', '\x77', '\x6f', '\x72', '\x6c', '\x64',
    '\x21', '\x00', '\x92', '\x00'
  };

  Hcode *object;

  object = inst_4th (EmbeddedHX);
  dump_4th (object, stdout, 0, -1);
  free_4th (object);
  return (EXIT_SUCCESS);
}

Input    : None
Output   : depends on program
Related  : load_4th()
*/

#ifdef USRLIB4TH
#include <4th.h>
#include <sys/cmds_4th.h>
#else
#include "4th.h"
#include "cmds_4th.h"
#endif

#include <stdlib.h>
#include <string.h>

static FILE*     HXIn;                  /* pointer to HX-file */
static unit*     BytecodeIn;            /* pointer to embedded HX */
static Hcode*    Object;                /* Hcode pointer */
static int       CSIn;                  /* checksum for reading */


#ifdef ARCHAIC
  static int (*LoadByte) ();           /* loader function (K&R) */
#else
  static int (*LoadByte) (void);       /* loader function (ANSI) */
#endif

/*
This function only sets ErrNo if no errors have occured yet.
*/

#ifndef ARCHAIC
  static void RaiseError (unsigned error)
#else
  static void RaiseError (error) unsigned error;
#endif

{
  if (! Object->ErrNo) Object->ErrNo = error;
}


/*
This low level function only loads a byte from the HX-file.
*/

#ifndef ARCHAIC
  static int LoadFileByte (void)
#else
  static int LoadFileByte ()
#endif

{
  int Hc;

  if ((Hc = fgetc (HXIn)) == EOF) RaiseError (M4IOERR);
  return (Hc);
}


/*
This low level function only loads a byte from the embedded HX-code.
*/

#ifndef ARCHAIC
  static int LoadEmbedByte (void)
#else
  static int LoadEmbedByte ()
#endif

{
  return (*(BytecodeIn++));
}


/*
This function only loads an addressunit from the HX-code if there have been
no errors up to now. Furthermore a checksum is calculated.
*/

#ifndef ARCHAIC
  static int LoadUnit (void)
#else
  static int LoadUnit ()
#endif

{
  int Hc;

  if (! Object->ErrNo)
     {
       Hc = (*(LoadByte)) ();
       CSIn ^= Hc;
       return (Hc);
     }

  return (0);
}


/*
This function loads a value in tiny, small or full length
*/

#ifndef ARCHAIC
  static cell LoadValue (size_t Max, char Sign)
#else
  static cell LoadValue (Max, Sign) size_t Max; char Sign;
#endif

{
  size_t n;                            /* number of bytes in a cell */
  cell Value = 0L;

  for (n = 0; n < (Max * HBITS); n += HBITS)
      Value += (((cell) LoadUnit ()) << n);

  return ((Sign ? -Value : Value));
}


/*
This function loads an integer value up to the size of a cell from an HX-file.
Like Forth, its components are broken up into an non-negative number and a
sign-byte.
*/

#ifndef ARCHAIC
  static cell LoadCell (void)
#else
  static cell LoadCell ()
#endif

{
  char Type  = (char) LoadUnit ();

  switch ((Type & (~HCSIGN)))    /* select the type of encoding */
    {
      case (HCZERO): return (Type & HSNEG ? CELL_MIN : 0L);
      case (HCONE ): return (Type & HSNEG ? -1L : 1L);
      case (HCBYTE): return (LoadValue (sizeof (char), Type & HSNEG));
      case (HCSHRT): return (LoadValue (sizeof (short), Type & HSNEG));
      case (HCCELL): return (LoadValue (sizeof (cell), Type & HSNEG));
      default      : RaiseError (M4BADOBJ); return (0L);
    }
}


/*
This function checks the checksum.
*/

#ifndef ARCHAIC
  static void LoadCheck (void)
#else
  static void LoadCheck ()
#endif

{
  int t = CSIn;

  if (t != LoadUnit ()) RaiseError (M4BADOBJ);
}


/*
This function checks whether the loaded executable is compatible with the
4tH virtual machine. If not, ErrNo4th is set.
*/

#ifndef ARCHAIC
  static void LoadInfo (void)
#else
  static void LoadInfo ()
#endif

{
  if (LoadUnit () == (char) sizeof (char))
     if (LoadUnit () == (char) sizeof (short))
        if (LoadUnit () == (char) sizeof (cell))
           if (LoadCell () == CELL_MAX)
              if (LoadCell () == (cell) Version4th)
                 if (LoadCell () == (cell) App4th) return;

  RaiseError (M4BADOBJ);
}


/*
This function creates and loads the Code Segment.
*/

#ifndef ARCHAIC
  static void LoadCode (void)
#else
  static void LoadCode ()
#endif

{
  int n;

  if (Object->ErrNo) Object->CodeSeg = NULL;
  else
    if ((Object->CodeSeg = (dict*) calloc ((size_t) Object->CodeSiz,
       sizeof (dict))) == NULL)
       RaiseError (M4NOMEM);
    else
       {
         for (n = 0; n < Object->CodeSiz; n++)
             {
               Object->CodeSeg [n].Word = (unit) LoadUnit ();
               switch (Object->CodeSeg [n].Word)
                 {
                   case (VARIABLE):
                   case (VALUE):
                   case (TO):
                   case (NOOP):
                   case (LITERAL):
                   case (PRINT):
                   case (SQUOTE):
                   case (BRANCH):
                   case (BRANCH0):
                   case (QDO):
                   case (LOOP):
                   case (PLOOP):
                   case (CALL): Object->CodeSeg [n].Value = LoadCell ();
                                break;
                   default:     Object->CodeSeg [n].Value = 0L;
                                break;
                 }
             }
       }
}


/*
This function creates and loads the String Segment.
*/

#ifndef ARCHAIC
  static void LoadString (void)
#else
  static void LoadString ()
#endif

{
  int n;

  if ((Object->ErrNo) || (Object->StringSiz == 0)) Object->StringSeg = NULL;
  else
    if ((Object->StringSeg = (char*) calloc ((size_t)
       Object->StringSiz, sizeof (char))) == NULL)
       RaiseError (M4NOMEM);
    else
       for (n = 0; n < Object->StringSiz; n++)
           Object->StringSeg [n] = (char) LoadUnit ();
}


/*
This routine creates, loads and initializes the header.
*/

#ifndef ARCHAIC
  static void LoadHeader (void)
#else
  static void LoadHeader ()
#endif

{
  Object->CodeSiz   = (int)      LoadCell ();
  Object->StringSiz = (unsigned) LoadCell ();
  Object->Variables = (unsigned) LoadCell ();
  Object->Strings   = (unsigned) LoadCell ();
}


/*
This is a generalized way to read in HX code, in whatever way it is stored
*/

#ifndef ARCHAIC
  static void LoadHX (void)
#else
  static void LoadHX ()
#endif

{
  if ((Object = (Hcode*) calloc (1, sizeof (Hcode))) == NULL) return;

  CSIn = 0;                            /* reset checksum */
  Object->ErrNo = M4NOERRS;            /* reset errors */

  LoadInfo   ();                       /* load version info */
  LoadHeader ();                       /* load header */
  LoadCode   ();                       /* load code segment */
  LoadString ();                       /* load string segment */
  LoadCheck  ();                       /* load checksum */
                                       /* on error: release all segments */
  if (Object->ErrNo)
    {
      if (Object->StringSeg) free (Object->StringSeg);
      if (Object->CodeSeg)   free (Object->CodeSeg);
      Object->CodeSiz  = Object->StringSiz = 0;
      Object->ErrLine  = 0;
      Object->Reliable = FALSE;
    }
  else                                 /* flag all ok */
    {
      Object->ErrLine  = Object->CodeSiz;
      Object->Reliable = TRUE;
    }
}


/*
The API hook to read HX files. It opens the file and sets up the proper
parameters. It returns a pointer to Hcode if successful, otherwise
a NULL pointer.
*/

#ifndef ARCHAIC
  Hcode* load_4th (char *filename)
#else
  Hcode* load_4th (filename) char* filename;
#endif

{
  if ((HXIn = fopen (filename, "rb")) == NULL) return (NULL);
  LoadByte = LoadFileByte;
  LoadHX ();
  if (fclose (HXIn)) RaiseError (M4IOERR);

  return (Object);
}


/*
The API hook to read embedded HX code. It sets up the proper
parameters. It returns a pointer to Hcode if successful, otherwise
a NULL pointer.
*/

#ifndef ARCHAIC
  Hcode* inst_4th (unit *embedded)
#else
  Hcode* inst_4th (embedded) unit *embedded;
#endif

{
  BytecodeIn = embedded;
  LoadByte = LoadEmbedByte;
  LoadHX ();
  return (Object);
}
