/*
+-------+-----------------+------+-----------------+-------+-----------------+
|Program| SAVE_4TH.C      |Header| 4TH.H           |Library| EXTLIB          |
+-------+-----------------+------+-----------------+-------+----------+------+
|HISTORY                                           |    Programmer    | Date |
+--------------------------------------------------+------------------+------+
| Turbo C V2.0 version                             | HanSoft          |210495|
| Linux V2.2.10 version                            | HanSoft          |190600|
+--------------------------------------------------+------------------+------+
|MODIFICATION LOG                                  |    Programmer    | Date |
+--------------------------------------------------+------------------+------+
| Added ErrLin4th                                  | HanSoft          |030595|
| Added support for Hcode                          | HanSoft          |120296|
| Added SaveBlock()                                | HanSoft          |130396|
| Removed SaveBlock(); added SaveByte(), SaveCell()| HanSoft          |100496|
| Added SaveInfo(), SaveHeader() and SaveCode()    | HanSoft          |100496|
| Added SaveString()                               | HanSoft          |100496|
| Rewrote in order to accomodate new Hcode format  | HanSoft          |021096|
| Renamed Appl4th to App4th                        | HanSoft          |031296|
| Added LOCAL_H support                            | HanSoft          |220597|
| Removed <stdio.h> and <limits.h>                 | HanSoft          |250797|
| Added BRANCHXT; changed header fieldname         | HanSoft          |120997|
| Removed BRANCHXT                                 | HanSoft          |170997|
| Added QDO                                        | HanSoft          |151097|
| Removed EasyC syntax                             | HanSoft          |150198|
| Checks open file before closing                  | HanSoft          |180298|
| Added SQUOTE                                     | HanSoft          |200298|
| Added SaveValue(); more compact HX format        | HanSoft          |140999|
| Cleaned up encoding constants                    | HanSoft          |150999|
| Fixed serious bug in SaveCell()                  | HanSoft          |111099|
| Always return CodeSiz in ErrLine                 | HanSoft          |031199|
| Fixed major bug; CSOut now properly initialized  | HanSoft          |111199|
| Fixed small error in K&R prototype               | HanSoft          |291299|
| Fixed some Linux warnings (no need; code is OK)  | HanSoft          |190600|
+--------------------------------------------------+------------------+------+
|REMARKS                                           |    Programmer    | Date |
+--------------------------------------------------+------------------+------+
| None                                             |                  |      |
+--------------------------------------------------+------------------+------+

  Copyright (C) 1997,2000 J.L. Bezemer

  This file is part of 4tH

  4tH is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


This routine writes the contents of an object to disk. It can then be loaded
with load_4th(). The object is pointed to by Program, filename must contain
a valid filename.

Prototype: void save_4th (Hcode *Program, char *filename)
Returns  : nothing
Memory   : none
Example  :


#include "4th.h"
#include <stdlib.h>

#ifndef ARCHAIC
  int main (int argc, char **argv)
#else
  int main (argc, argv) int argc; char **argv;
#endif

{
  Hcode *object;
  char  *sources;

  if (argc == 2)
    {
      if ((sources = open_4th (argv [1])) == NULL)
         printf ("Loading; \'%s\' does not exist or too large\n", argv [1]);
      else
        {
          object = comp_4th (sources);
          if (object)
              printf ("Compile; word %u: %s\n", object->ErrLine,
                     errs_4th [object->ErrNo]);
          save_4th (object, "out.hx");
          return (object->ErrNo ? EXIT_FAILURE : EXIT_SUCCESS);
        }
    }
  return (EXIT_FAILURE);
}


Input    : None
Output   : depends on program
Related  : dump_4th(), exec_4th()
*/

#ifdef USRLIB4TH
#include <4th.h>
#include <sys/cmds_4th.h>
#else
#include "4th.h"
#include "cmds_4th.h"
#endif

#define SaveCheck() SaveByte(CSOut)

static FILE*     HXOut;                /* pointer to HX-file */
static int       CSOut;                /* checksum */
static unsigned* SaveErr;

/*
This function only saves a byte to the HX-file if there have been no errors up
to now. Furthermore a checksum is generated. In fact a dedicated version of
fputc().
*/

#ifndef ARCHAIC
  static void SaveByte (int Value)
#else
  static void SaveByte (Value) int Value;
#endif

{
  if (! *SaveErr)
     {
       if (fputc (Value, HXOut) == EOF) *SaveErr = M4IOERR;
       else CSOut ^= Value;
     }
}


/*
This function saves a value in tiny, small or full length.
*/

#ifndef ARCHAIC
  static void SaveValue (cell Value, size_t Max)
#else
  static void SaveValue (Value, Max) cell Value; size_t Max;
#endif

{

  size_t n;

  for (n = 0 ; n < Max; n++)
      {                                /* save each byte */
        SaveByte ((int) (Value & (cell) HMASK));
        Value >>= HBITS;               /* get the next byte */
      }
}


/*
This function saves an integer value up to the size of a cell to an HX-file.
Like Forth, its components are broken up into an non-negative number and a
sign-byte.
*/

#ifndef ARCHAIC
  static void SaveCell (cell Value)
#else
  static void SaveCell (Value) cell Value;
#endif

{
  int Sign;
                                       /* handle special cases first */
      if      (Value == -1L)      SaveByte (HCONE  | HSNEG);
      else if (Value ==  0L)      SaveByte (HCZERO | HSPOS);
      else if (Value ==  1L)      SaveByte (HCONE  | HSPOS);
      else if (Value == CELL_MIN) SaveByte (HCZERO | HSNEG);
      else {
             if ((Sign = (Value < 0 ? HSNEG : HSPOS)) == TRUE) Value = -Value;
                                       /* take care of the sign */
             if (Value <= (cell) HNTINY)
                {                      /* this fits in a byte */
                   SaveByte (HCBYTE | Sign);
                   SaveValue (Value, sizeof (char));
                }
             else if (Value <= (cell) HNSMALL)
                {                      /* this fits in a short */
                   SaveByte (HCSHRT | Sign);
                   SaveValue (Value, sizeof (short));
                }
             else
                {                      /* this number needs full cell */
                   SaveByte (HCCELL | Sign);
                   SaveValue (Value, sizeof (cell));
                }
      }
}


/*
This function saves all info which identifies the HX-program.
*/

#ifndef ARCHAIC
  static void SaveInfo (void)
#else
  static void SaveInfo ()
#endif

{
  SaveByte (sizeof (char));
  SaveByte (sizeof (short));
  SaveByte (sizeof (cell));
  SaveCell (CELL_MAX);                 /* save sizes of numbers */
  SaveCell ((cell) Version4th);        /* save version info */
  SaveCell ((cell) App4th);
}


/*
This function saves all information which describes the programs runtime-
environment.
*/

#ifndef ARCHAIC
  static void SaveHeader (Hcode* ObjHead)
#else
  static void SaveHeader (ObjHead) Hcode* ObjHead;
#endif

{
  SaveCell ((cell) ObjHead->CodeSiz);
  SaveCell ((cell) ObjHead->StringSiz);
  SaveCell ((cell) ObjHead->Variables);
  SaveCell ((cell) ObjHead->Strings);
}


/*
This function saves the program-code in a condensed format. Only those
words which need an argument have their arguments saved.
*/

#ifndef ARCHAIC
  static void SaveCode (Hcode* ObjCode)
#else
  static void SaveCode (ObjCode) Hcode* ObjCode;
#endif

{
  int n;

  for (n = 0; n < ObjCode->CodeSiz; n++)
      {
        SaveByte ((int) ObjCode->CodeSeg [n].Word);

        switch (ObjCode->CodeSeg [n].Word)
          {
            case (VARIABLE):
            case (VALUE):
            case (TO):
            case (NOOP):
            case (LITERAL):
            case (PRINT):
            case (SQUOTE):
            case (BRANCH):
            case (BRANCH0):
            case (QDO):
            case (LOOP):
            case (PLOOP):
            case (CALL): SaveCell (ObjCode->CodeSeg [n].Value); break;
          }
      }
}


/*
This function simply saves the contents of the StringSegment.
*/

#ifndef ARCHAIC
  static void SaveString (Hcode* ObjStr)
#else
  static void SaveString (ObjStr) Hcode* ObjStr;
#endif

{
  int n;

  for (n = 0; n < ObjStr->StringSiz; n++)
      SaveByte ((int) ObjStr->StringSeg [n]);
}


#ifndef ARCHAIC
  void save_4th (Hcode* Object, char* filename)
#else
  void save_4th (Object, filename) Hcode* Object; char* filename;
#endif

{
  if (! Object) return;              /* check if object is reliable */
  else if (! Object->Reliable) return;

  CSOut           = 0;               /* reset checksum */
  SaveErr         = &(Object->ErrNo);
  Object->ErrNo   = M4NOERRS;
  Object->ErrLine = Object->CodeSiz;

  if ((HXOut = fopen (filename, "wb")) == NULL) Object->ErrNo = M4IOERR;

  SaveInfo   ();
  SaveHeader (Object);
  SaveCode   (Object);
  SaveString (Object);
  SaveCheck  ();

  if (HXOut) if (fclose (HXOut)) Object->ErrNo = M4IOERR;
}
